"use strict";
this.name	= "trails";
this.author	= "Norby";
this.copyright	= "2015 Norby";
this.description= "Ships leaves condensation trails";
this.licence	= "CC BY-NC-SA 4.0";

this.$MaxKey = 3; //max. used dataKey from effectdata.plist ("trails3"), 4 need much power
this.$MinFPS = 25; //if fall below then will reduce the max. dataKey number

//internal properties, should not touch
this.$Delta = 0; //delta counter for FCB
this.$Delta2 = 0; //store first delta in the last second for FCB
this.$Distance = 40; //distance between trail elements based on effectada.plist
this.$ED = null; //store a pointer to escortdeck worldscript
this.$FC = 0; //frame counter for calculating FPS
this.$FP = null; //FCB pointer
this.$FPS = 60; //lastly measured FPS
this.$Key = this.$MaxKey; //dataKey in effectdata ("trails3")
this.$Length = this.$MaxLength; //actual length to avoid "Universe is full"
this.$MaxLength = 50; //how many elements are in a trail (40m each) based on effectdata.plist
this.$Old = false; //old version before Oolite v1.82
this.$S = [];	//tracked ships and visualEffects
		//$S[i][0]: the ship object,
		//$S[i][1]: last position of the ship where a visualEffect is created
		//$S[i][2]: lastly used "k" indexes
                //$S[i][k] where k >= 3: visualEffects
//this.$T = [];	//timestamps of the creation of the last trail elements
//this.$TK = [];//"k" indexes of timestamped last trail elements
//this.$V = [];	//storage of all visualEffects

//worldscript events
this.startUp = function() {
	this.$ED = worldScripts.escortdeck;
	this.$Clear(this);
	this.$FC = 0; //frame counter
	this.$Delta = 0; //delta counter
	if (0 < oolite.compareVersion("1.82")) {
		this.$Old = true;
/*		
		//pre-creating visualEffects
		var key = "trails";
		var pos = Vector3D(100000000,0,0);
		for( var i = 0; i < 100; i++ ) {
			this.$V.push(system.addVisualEffect( key, pos ));
		}
*/		
		if( !isValidFrameCallback( this.$FP ) )
			this.$FP = addFrameCallback( this.$FCBOld.bind(this) );
	} else { //from Oolite v1.82 flashers in subentities are usable
		this.$Old = false;
		if( !isValidFrameCallback( this.$FP ) )
			this.$FP = addFrameCallback( this.$FCB.bind(this) );
	}
}

this.shipWillEnterWitchspace = function() {
	this.$Clear(this);
}

//local methods
this.$Clear = function(w) {  //start a new array, free up memory
	w.$Length = w.$MaxLength;
	delete w.$S;
	w.$S = [];
	w.$S[0] = [];
	w.$S[0][0] = player.ship; //the first ship
}

this.$FCB = function( delta ) {
	var w = this; //worldscript
	var p = player.ship;
	var pp = false;
	if( p && p.isValid ) pp = p.position;
	w.$FC++; //frame counter for calculating FPS
	w.$Delta += delta;
	if( w.$Delta > w.$Delta2 + 1 && pp ) { //like an 1s timer, search for new ships
		w.$FPS = w.$FC; //save measured fps for the trail length reducer code below
		w.$FC = 0; //reset frame counter
		w.$Delta2 = w.$Delta; //save delta
		var ships = [];
		if( !p.docked ) ships = p.checkScanner(false);//faster than filteredEntities
				//must use checkScanner with false to add the "green" escorts
				//and a good side effect boulders get steam tails
				//moreover true return the full list in Oolite 1.80 (fixed in 1.81)
		for( var i = 1; i < w.$S.length; i++ ) { //remove trails of far ships
			var o = w.$S[i]; //data of an old ship
			var x = -2;
			if( o && o[0] && o[0].isValid ) {
				x = ships.indexOf( o[0] );
				if( w.$ED ) { //disable escort trails on deck, enable if launched
					if( !o[0].script ) //for sure
						o[0].setScript("oolite-default-ship-script.js");
					var pad = w.$ED.$EscortDeckShip.indexOf( o[0] );
					if( pad != -1 && w.$ED.$EscortDeckShipPos[pad] ) {
						o[0].script.$TrailsDisabled = true;
					} else o[0].script.$TrailsDisabled = false;
				}
			}
			//if invalid then docked, jumped or destroyed, leave to run remove below
			if( x == -1 ) { //not in the scanner anymore?
				//must check distance due to many ships near exhibitions
				//can leave out important ones from the limited checkScanner list
				if( pp.distanceTo( o[0].position ) > 2 * p.scannerRange ) {
					for( var k = 3; k <= w.$Length; k++ ) {
						var v = o[k]; //remove visualEffects
						if( v && v.isValid ) v.remove();
					}
					w.$S[i] = false; //remove this ship from the tracked array
				}
			} else if( x >= 0 ) ships[x] = false; //don't add into $S if already in

		}
		
		if( p.docked ) {
			if( w.$S[1] ) w.$Clear(w); //start a new array, free up memory
		} else for( var i = 0; i < ships.length; i++ ) { //add new ships into $S
			var ship = ships[i];
			if( ship && ship.isValid && ship.maxSpeed > 0 //do not add buoy, etc.
				&& ( ship.exhausts && ship.exhausts[0] //need valid exhaust
					|| ship.scanClass == "CLASS_ROCK" )
				&& ship.dataKey != "telescopemarker" //exclude for Telescope
				&& ship.scanClass != "CLASS_NO_DRAW" ) { //exclude for sure
				//do not add escorts landed on EscortDeck
				var add = true;
				if( w.$ED ) {
					var pad = w.$ED.$EscortDeckShip.indexOf( ship );
					if( pad != -1 && w.$ED.$EscortDeckShipPos[pad] ) {
						add = false;
					}
				}
				if( add ) {
					var l = w.$S.length; //add after the last ship
					w.$S[l] = [];
					w.$S[l][0] = ship;
				}
			}
		}
	}

	//reduce the number of flashers if FPS is low
	var newkey = false;
	if( w.$FPS < w.$MinFPS ) {
		w.$FPS = w.$MinFPS; //do not reduce again until the next measue
		if( w.$Key > 2 ) { //1 is valid but not nice
			w.$Key--;
			newkey = true;
		}
	} else if( w.$FPS > 2 * w.$MinFPS ) { //extend up to max.length
		w.$FPS = 2 * w.$MinFPS; //do not extend again until the next measue
		if( w.$Key < w.$MaxKey ) {
			w.$Key++;
			newkey = true;
		}
	}
	var key = "trails"+w.$Key;
	if( newkey ) for( var i = 0; i < w.$S.length; i++ ) { //replace all trails
                var ship = w.$S[i][0];
                if( ship.scanClass == "CLASS_ROCK" ) continue;//key = "trails-rock";
                else if( w.$Key > 1 && ship.primaryRole == "wreckage" //speedup for explosions
                        || ship.primaryRole == "oolite-wreckage-chunk" ) continue;//key = "trails1";
                else key = "trails"+w.$Key;
		var t = w.$S[i], z;
		if( t ) {
			for( var k = 3; k <= w.$Length; k++ ) {
				var v = t[k];
				if( v && v.isValid ) {
					z = system.addVisualEffect(key,v.position);
					if( z ) {
						z.orientation=v.orientation;
						v.remove();
						w.$S[i][k] = z;
					}
				}
			}
		}
	}

	//reduce length over 1000 entity to prevent "Universe is full" error at 2047
	var a = system.allShips.length + system.allVisualEffects.length;
	if( a > 1000 && w.$Length > 10 ) { //low fps or too much entity
		var r = Math.ceil( w.$Length/2 ); //reduced length
		for( var i = 0; i < w.$S.length; i++ ) {
			var t = w.$S[i];
			if( t ) {
				//remove the whole long trail
				for( k = 3; k <= w.$Length; k++ ) {
					var v = t[k];
					if( v && v.isValid ) v.remove();
				}
				w.$S[i][2] = 3;//reset visualEffect index
			}
		}
		w.$Length = r;
	}
	
	//create new trail elements
	var dist = w.$Distance;
	for( var i = 1; i < w.$S.length; i++ ) { //start from 0 and player ship will make trail also
		var t = w.$S[i];
		if( t ) {
			var ship = t[0];
			if( ship && ship.isValid && !ship.isDerelict
				&& !ship.script.$TrailsDisabled ) {
				var sp = ship.position;
				var sc = 1;
				
                                if( ship.scanClass == "CLASS_ROCK" ) key = "trails-rock";
                                else if( w.$Key > 1 && ship.primaryRole == "wreckage" //speedup for explosions
                                        || ship.primaryRole == "oolite-wreckage-chunk" ) key = "trails1";
                                else {
                                        key = "trails"+w.$Key;//need if the previous ship was over 10km
				        if( pp ) { //create less flashers over 12km
					        var pd = sp.distanceTo( pp ) / 1000;
					        if( pd > 12 ) {
						        key = "trails1";
						        sc = 1 + Math.max(0, Math.min( 4, pd/5-2 ));
					        }
				        }	//increase size from 12km to 25km to keep up brightness
                                }

				var c = 0; //safety counter to surely exit from the next cycle soon
				//check distance of the last position where a trail is created
				//need to do in cycle if travelled more in the last frame than
				//the length of a visualEffect element in effectdata.plist (20m)
				//which is happen often with more than 1x TAF
				var k = t[2];
				while( c < 100 && ( !k || !w.$S[i][k] || !w.$S[i][k].isValid
					|| sp.distanceTo( w.$S[i][k].position ) > 2*dist ) ) {
					c++;
					if( !w.$S[i][2] ) w.$S[i][2] = 3; //new trail
					k = w.$S[i][2];
					if( !k ) k = 3;
					var a = w.$S[i][k]; //actual (lastly created) visualEffect
					var pos = sp;
					var q = ship.orientation; //for the first element
					if( a && a.isValid ) { //from the second element
						var u = sp.subtract(a.position).direction();
						pos = a.position.add(u.multiply(dist));
						if( a && a.isVisualEffect ) {
							//facing from the previous trail element
							q = a.orientation;
							var z=pos.subtract(a.position).direction();
							var ah = a.heading;
							var angle = ah.angleTo(z);
							//set the plane where we should rotate in
							var cr = ah.cross(z).direction();
							q = q.rotate( cr, -angle );
						}
					}
					w.$S[i][1] = pos; //save the new position - unused

					k++; //index of the next item
					if( k > w.$Length ) k = 3; //restart over $Length
					w.$S[i][2] = k; //save the new index
					
                                        //make the end of the trail darker but prevent blue end on gray dust
					if( ship.scanClass != "CLASS_ROCK" ) {
					        var pk = k+1; //previous k points to the oldest visualEffect
					        if( pk > w.$Length ) pk = 3;
					        var pv = w.$S[i][pk]; //the oldest visualEffect
					        if( pv && pv.isVisualEffect ) {
						        var nv = system.addVisualEffect(
							        "trails-end", pv.position );
						        if( nv ) {
							        nv.orientation = pv.orientation;
							        pv.remove();
							        w.$S[i][pk] = nv; //save the new element
                                                        }
						}
					}

					//replace old visual item
					var ov = w.$S[i][k];
					if( ov && ov.isVisualEffect ) ov.remove();
					var v = system.addVisualEffect( key, pos );
					if( v ) {
						v.orientation = q;
						w.$S[i][k] = v; //save the new element
					}

				}
				//increase flasher size with the distance
				if( sc > 1 ) for( var j = 3; j <= w.$Length; j++ ) {
					var v = w.$S[i][j];
					if( v && v.isValid ) v.scaleX = v.scaleY = sc;
				}				//but not scaleZ!
			} else { //ship docked, jumped or destroyed
				var k = t[2]; //actual item index - old code without timestamp
				if( k ) {
					k++; //+= 0.5; //slower count for slower remove
					if( k > w.$Length ) k = 3; //restart over $Length
					w.$S[i][2] = k; //save the new index
					if( k == Math.floor(k) ) {
						if( w.$S[i][k] == -1 ) { //all visualEffect removed
							//so remove this ship from the tracked array
							for( var k = 3; k <= w.$Length; k++ ) {
								var v = t[k]; //remove again for sure
								if( v && v.isValid ) v.remove();
							}
							w.$S[i] = false; 
						} else {
							var ov = t[k]; //remove old visual item
							if( ov && ov.isVisualEffect ) ov.remove();
							w.$S[i][k] = -1; //flag for remove ship
						}
					}
				}
			}

/*			//this part is not working yet, need more debugging
			//verify timestamp and remove the last element after max. 0.1s
			if( w.$Delta > w.$T[i] + 0.1 ) {
				var k = w.$TK[i];
				var tv = w.$S[i][k];
				if( tv && tv.isVisualEffect ) tv.remove();
				w.$S[i][k] = false;
				
				//replace the next element to the dark one
				if( t[2] != w.$TK[i] ) { //do not step over the first trail element
					w.$TrailEnd( w, i, w.$TK[i]+1 );
				}
			}
*/		}
	}
}


this.$FCBOld = function( delta ) { //before Oolite 1.82
	var w = this; //worldscript
	var p = player.ship;
	var pp = false;
	if( p && p.isValid ) pp = p.position;
	w.$FC++; //frame counter for calculating FPS
	w.$Delta += delta;
/*	log(w.name, "FC:"+w.$FC+" delta:"+delta);
	if( p.target && p.target.isBeacon ) { //speedtest
		var pos = Vector3D(100000000000,100000000000,100000000000);
		var sv = [];
		for( var i = 1; i < 100; i++ )
			sv.push(system.addVisualEffect( "trails1", pos ));
		for( var i = 1; i < 100; i++ )
			if(sv[i]) sv[i].remove();
	}*/

	if( w.$Delta > w.$Delta2 + 1 && pp ) { //like an 1s timer, search for new ships
		w.$FPS = w.$FC; //save measured fps for the trail length reducer code below
		w.$FC = 0; //reset frame counter
		w.$Delta2 = w.$Delta; //save delta
		var ships = [];
		if( !p.docked ) ships = p.checkScanner(false);//faster than filteredEntities
				//must use checkScanner with false to add the "green" escorts
				//and a good side effect boulders get steam tails
				//moreover true return the full list in Oolite 1.80 (fixed in 1.81)
		for( var i = 1; i < w.$S.length; i++ ) { //remove trails of far ships
			var o = w.$S[i]; //data of an old ship
			var x = -2;
			if( o && o[0] && o[0].isValid ) {
				x = ships.indexOf( o[0] );
				if( w.$ED ) { //disable escort trails on deck, enable if launched
					if( !o[0].script ) //for sure
						o[0].setScript("oolite-default-ship-script.js");
					var pad = w.$ED.$EscortDeckShip.indexOf( o[0] );
					if( pad != -1 && w.$ED.$EscortDeckShipPos[pad] ) {
						o[0].script.$TrailsDisabled = true;
					} else o[0].script.$TrailsDisabled = false;
				}
			}
			//if invalid then docked, jumped or destroyed, leave to run remove below
			if( x == -1 ) { //not in the scanner anymore?
				//must check distance due to many ships near exhibitions
				//can leave out important ones from the limited checkScanner list
				if( pp.distanceTo( o[0].position ) > 2 * p.scannerRange ) {
					for( var k = 3; k <= w.$Length; k++ ) {
						var v = o[k]; //remove visualEffects
						if( v && v.isValid ) v.remove();
					}
					w.$S[i] = false; //remove this ship from the tracked array
				}
			} else if( x >= 0 ) ships[x] = false; //don't add into $S if already in

		}
		
		if( p.docked ) {
			if( w.$S[1] ) w.$Clear(w); //start a new array, free up memory
		} else for( var i = 0; i < ships.length; i++ ) { //add new ships into $S
			var n = ships[i];
			if( n && n.isValid && n.maxSpeed > 0 //do not add buoy, etc.
				&& ( n.exhausts && n.exhausts[0] //need valid exhaust
					|| n.scanClass == "CLASS_ROCK" ) //steam tails
				&& n.dataKey != "telescopemarker" //exclude for Telescope
				&& n.scanClass != "CLASS_NO_DRAW" ) { //exclude for sure
				//do not add escorts landed on EscortDeck
				var add = true;
				if( w.$ED ) {
					var pad = w.$ED.$EscortDeckShip.indexOf( n );
					if( pad != -1 && w.$ED.$EscortDeckShipPos[pad] ) {
						add = false;
					}
				}
				if( add ) {
					var l = w.$S.length; //add after the last ship
					w.$S[l] = [];
					w.$S[l][0] = n;
				}
			}
		}
		
//		if(pp) system.addVisualEffect( "trails3-", pp ); //debug
	}

	//reduce the number of flashers if FPS is low
	var newkey = false;
	if( w.$FPS < w.$MinFPS ) {
		w.$FPS = w.$MinFPS; //do not reduce again until the next measue
		if( w.$Key > 1 ) {
			w.$Key--;
			newkey = true;
		}
	} else if( w.$FPS > 2 * w.$MinFPS ) { //extend up to max.length
		w.$FPS = 2 * w.$MinFPS; //do not extend again until the next measue
		if( w.$Key < w.$MaxKey ) {
			w.$Key++;
			newkey = true;
		}
	}
	var key = "trails"+w.$Key;
	if( newkey ) for( var i = 0; i < w.$S.length; i++ ) { //replace all trails
		var t = w.$S[i], z;
		if( t ) {
			for( var k = 3; k <= w.$Length; k++ ) {
				var v = t[k];
				if( v && v.isValid ) {
					z = system.addVisualEffect(key,v.position);
					if( z ) {
						z.orientation=v.orientation;
						v.remove();
						w.$S[i][k] = z;
					}
				}
			}
		}
	}
//log(w.name, "wkey:"+w.$Key+" key:"+key);

	//reduce length over 1000 entity to prevent "Universe is full" error at 2047
	var a = system.allShips.length + system.allVisualEffects.length;
	if( a > 1000 && w.$Length > 10 ) { //low fps or too much entity
		var r = Math.ceil( w.$Length/2 ); //reduced length
		for( var i = 0; i < w.$S.length; i++ ) {
			var t = w.$S[i];
			if( t ) {
/*				//remove only the second half of trails - not enough good
//log(w.name, "i:"+i+" t:"+t);
				var tmp = [];
				var j = 3;
				var ka = w.$S[i][2];
				var k = ka;
				for( k = k; k <= w.$Length && j <= r; k++ ) {
					tmp[j++] = t[k]; //save the first half
				}
//log(w.name, "ka:"+ka+" k:"+k+" j:"+j+" r:"+r+" tmp.length:"+tmp.length);
				if( k >= t.length ) { //continue from the begin of array until r
					for( k = 3; k < ka && k <= w.$Length && j <= r; k++ ) {
						tmp[j++] = t[k];
					}
				} else { //remove visualEffects until the end of array
					for( k = k; k <= w.$Length; k++ ) {
						var v = t[k];
						if( v && v.isValid ) v.remove();
					}
				}
//log(w.name, "k:"+k+" j:"+j+" r:"+r+" tmp.length:"+tmp.length);
				for( k = 3; k < ka && k <= w.$Length; k++ ) {
					var v = t[k]; //remove visualEffects from the begin of array
					if( v && v.isValid ) v.remove();
				}
				for( k = 3; k <= r; k++ ) {
					w.$S[i][k] = tmp[k];
				}
//log(w.name, "s.length:"+w.$S[i].length);
*/
				//remove the whole long trail
				for( k = 3; k <= w.$Length; k++ ) {
					var v = t[k];
					if( v && v.isValid ) v.remove();
				}
/*
				var e = t[t[2]];
				if( e && e.isVisualEffect ) {
					if( e.dataKey != "trails2" ) {
						for( var k = 3; k <= w.$Length; k++ ) {
							var v = t[k];
							if( v && v.isValid ) {
								var z = system.addVisualEffect(
									"trails2", v.position );
								if( z ) {
									z.orientation=v.orientation;
									v.remove();
									w.$S[i][k] = z;
								}
							}
						}
					}
				}
*/
				w.$S[i][2] = 3;//reset visualEffect index
			}
		}
		w.$Length = r;
	} //else if( a < 500 )  //extend up to max.length
	//	w.$Length = Math.min( 2 * w.$Length, w.$MaxLength );

	//create new trail elements
	var dist = w.$Distance;
//	var sumc = 0;
	for( var i = 1; i < w.$S.length; i++ ) { //start from 0 and player ship will make trail also
		var t = w.$S[i];
		if( t ) {
			var ship = t[0];
			if( ship && ship.isValid && !ship.isDerelict
				&& !ship.script.$TrailsDisabled ) {
				var sp = ship.position;
				var sc = 1;
				key = "trails"+w.$Key; //need if the previous ship was over 15km
				if( pp ) { //create less flashers over 12km
					var pd = sp.distanceTo( pp ) / 1000;
					if( pd > 12 ) {
						key = "trails1";
						sc = 1 + Math.max(0, Math.min( 4, pd/5-2 ));
					}
				}	//increase size from 12km to 25km to keep up brightness

//				var ed = 0; -removed to prevent a bug
//				if( ship.exhausts && ship.exhausts[0] && ship.exhausts[0].position )
//					ed = ship.exhausts[0].position.z;
				var c = 0; //safety counter to surely exit from the next cycle soon
				//check distance of the last position where a trail is created
				//need to do in cycle if travelled more in the last frame than
				//the length of a visualEffect element in effectdata.plist (20m)
				//which is happen often with more than 1x TAF
				var k = t[2];
				while( c < 100 && ( !k || !w.$S[i][k] || !w.$S[i][k].isValid
					|| sp.distanceTo( w.$S[i][k].position ) > 2*dist ) ) {
				//if(ship == p.target) log(w.name, "i:"+i+" c:"+c+" dist:"+dist);
					c++;
//if(ship == p.target) log(w.name, "wk:"+w.$Key+" key:"+key+" pd:"+Math.round(pd));
					if( !w.$S[i][2] ) w.$S[i][2] = 3; //new trail
					k = w.$S[i][2];
					if( !k ) k = 3;
					var a = w.$S[i][k]; //actual (lastly created) visualEffect
					var pos = sp;
					var q = ship.orientation; //for the first element
					if( a && a.isValid ) { //from the second element
						var u = sp.subtract(a.position).direction();
						pos = a.position.add(u.multiply(dist));
						if( a && a.isVisualEffect ) {
							//facing from the previous trail element
							q = a.orientation;
							var z=pos.subtract(a.position).direction();
							var ah = a.heading;
							var angle = ah.angleTo(z);
							//set the plane where we should rotate in
							var cr = ah.cross(z).direction();
							q = q.rotate( cr, -angle );
						}
					}
					w.$S[i][1] = pos; //save the new position - unused

					k++; //index of the next item
					if( k > w.$Length ) k = 3; //restart over $Length
					w.$S[i][2] = k; //save the new index
					
					//make the end of the trail darker if length is maximal
//					if( k == w.$TK[i] ) w.$TrailEnd( w, i, k+1 );

					//make the end of the trail darker
					var pk = k+1; //previous k points to the oldest visualEffect
					if( pk > w.$Length ) pk = 3;
					var pv = w.$S[i][pk]; //the oldest visualEffect
					if( pv && pv.isVisualEffect ) {
						var nv = system.addVisualEffect(
							"trails-end", pv.position );
						if( nv ) {
							nv.orientation = pv.orientation;
							pv.remove();
							w.$S[i][pk] = nv; //save the new element
						}
					}

					//replace old visual item
					var ov = w.$S[i][k];
					if( ov && ov.isVisualEffect ) ov.remove();
					var v = system.addVisualEffect( key, pos );
					if( v ) {
						v.orientation = q;
						w.$S[i][k] = v; //save the new element
					}

				}
//				sumc += c;
//if(ship == p.target) log(w.name, " c:"+c+" dist:"+dist+" dto:"+sp.distanceTo( w.$S[i][k].position ));
				
/*				//replace to less flasher over 5km - cause momentary FPS drops
				var e = w.$S[i][w.$S[i][2]];
				if( e && e.isVisualEffect && pp && w.$Key > 1 ) {
					if( e.dataKey != key ) { //replace to more or less flasher
						for( var k = 3; k <= w.$Length; k++ ) {
							var v = w.$S[i][k];
							if( v && v.isValid ) {
								var z = system.addVisualEffect(
									key, v.position );
								if( z ) {
									z.orientation=v.orientation;
									v.remove();
									w.$S[i][k] = z;
								}
							}
						}
					}
				}

*/
				//increase flasher size with the distance
				if( sc > 1 ) for( var j = 3; j <= w.$Length; j++ ) {
					var v = w.$S[i][j];
					if( v && v.isValid ) v.scaleX = v.scaleY = sc;
				}				//but not scaleZ!

/*				//small size (10%) at the end of trail - bad idea
				var j = k;
				while( j++ < w.$Length ) {
					var v = t[j];
					var sc = 0.1 + 0.9 * ( w.$Length - (j - k) ) / w.$Length;
					if( v && v.isValid )
						v.scaleX = v.scaleY = sc;//but not scaleZ!
				}
				j = 3;
				while( j++ < k ) {
					var v = t[j];
					var sc = 0.1 + 0.9 * ( k - j + 2 ) / w.$Length;
					if( v && v.isValid )
						v.scaleX = v.scaleY = sc;//but not scaleZ!
				}
*/
/*				//this part is not working yet, need more debugging
				if( pp ) { //increase flasher size with the distance from the player
					var pd = sp.distanceTo( pp ) / 1000;
					var sc = 1 + Math.max( 0, pd - 3 );
					for( var k = 3; k <= w.$Length; k++ ) {
						var v = t[k];
						if( v && v.isValid )
							v.scaleX = v.scaleY = sc;//but not scaleZ!
					}
				}
*/			} else { //ship docked, jumped or destroyed
/*				if( t[2] == w.$TK ) { //all visualEffect expired
					for( var k = 3; k <= w.$Length; k++ ) {
						var v = t[k]; //remove again for sure
						if( v && v.isValid ) v.remove();
					}
					//remove this ship from the tracked array
					w.$S[i] = false; 
				}
*/				var k = t[2]; //actual item index - old code without timestamp
				if( k ) {
					k++; //+= 0.5; //slower count for slower remove
					if( k > w.$Length ) k = 3; //restart over $Length
					w.$S[i][2] = k; //save the new index
					if( k == Math.floor(k) ) {
						if( w.$S[i][k] == -1 ) { //all visualEffect removed
							//so remove this ship from the tracked array
							for( var k = 3; k <= w.$Length; k++ ) {
								var v = t[k]; //remove again for sure
								if( v && v.isValid ) v.remove();
							}
							w.$S[i] = false; 
						} else {
							var ov = t[k]; //remove old visual item
							if( ov && ov.isVisualEffect ) ov.remove();
							w.$S[i][k] = -1; //flag for remove ship
						}
					}
				}
			}

/*			//this part is not working yet, need more debugging
			//verify timestamp and remove the last element after max. 0.1s
			if( w.$Delta > w.$T[i] + 0.1 ) {
				var k = w.$TK[i];
				var tv = w.$S[i][k];
				if( tv && tv.isVisualEffect ) tv.remove();
				w.$S[i][k] = false;
				
				//replace the next element to the dark one
				if( t[2] != w.$TK[i] ) { //do not step over the first trail element
					w.$TrailEnd( w, i, w.$TK[i]+1 );
				}
			}
*/		}
	}
//	if( ships ) c = c + " ships:"+ships.length;
//	log(w.name, "i:"+i+" sumc:"+sumc);
}
/*
this.$TrailEnd = function( w, i, tk ) { //create new dark ending element of a trail
	var ov = w.$S[i][w.$TK[i]]; //remove the old "trails-end" visualEffect
	if( ov && ov.isVisualEffect ) ov.remove();
	w.$S[i][w.$TK[i]] = false;
	
	if( tk > w.$Length ) tk = 3;
	var pv = w.$S[i][tk]; //replace the oldest "trails" visualEffect
	var nv = system.addVisualEffect( "trails-end", pv.position );
	if( pv && pv.isVisualEffect ) {
		if( nv ) nv.orientation = pv.orientation;
		pv.remove();
	}
	w.$S[i][tk] = nv; //save the new element
	
	w.$T[i] = w.$Delta; //set the timestamp of creation
	w.$TK[i] = tk; //save the "k" index of the timestamped last element
}
*/